/**
  *  \file recog_mp3.c
  *  \brief Recognizing MP3 Files
  *
  *  Much of this stolen from mpg123.
  */

#include "ringbuf.h"
#include "recog.h"
#include "output.h"

static int tabsel_123[2][3][16] = {
    { { 0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,},
      { 0,32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384,},
      { 0,32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320,} },
       
    { { 0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,},
      { 0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,},
      { 0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,} }
};

static long freqs[9] = { 44100, 48000, 32000, 22050, 24000, 16000 , 11025 , 12000 , 8000 };

/** true iff head is plausible (=accepted by mpg123) */
static int
head_check(unsigned long head)
{
    if( (head & 0xffe00000) != 0xffe00000)
	return 0;
    if(!((head>>17)&3))
	return 0;
    if( ((head>>12)&0xf) == 0xf)
	return 0;
    if( ((head>>10)&0x3) == 0x3 )
	return 0;
    if ((head & 0xffff0000) == 0xfffe0000)
        return 0;

    return 1;
}

static long
decode_header(unsigned long newhead)
{
    int lsf, mpeg25, lay, sampling_frequency, bitrate_index, padding;
    long framesize;
    
    if(!head_check(newhead))
        return 0;

    if( newhead & (1<<20) ) {
        lsf = (newhead & (1<<19)) ? 0x0 : 0x1;
        mpeg25 = 0;
    }
    else {
        lsf = 1;
        mpeg25 = 1;
    }
    
    lay = 4-((newhead>>17)&3);
    if( ((newhead>>10)&0x3) == 0x3)
        return 0;               /* "Stream error" */
    if(mpeg25)
        sampling_frequency = 6 + ((newhead>>10)&0x3);
    else
        sampling_frequency = ((newhead>>10)&0x3) + (lsf*3);

    bitrate_index = ((newhead>>12)&0xf);
    padding   = ((newhead>>9)&0x1);

    if(!bitrate_index)
        return 0;               /* "Free format not supported" */
    if (lay != 3)
        return 0;

    switch(lay) {
     case 1:
        framesize  = (long) tabsel_123[lsf][0][bitrate_index] * 12000;
        framesize /= freqs[sampling_frequency];
        framesize  = ((framesize+padding)<<2)-4;
        break;
     case 2:
        framesize = (long) tabsel_123[lsf][1][bitrate_index] * 144000;
        framesize /= freqs[sampling_frequency];
        framesize += padding - 4;
        break;
     case 3:
        framesize  = (long) tabsel_123[lsf][2][bitrate_index] * 144000;
        framesize /= freqs[sampling_frequency]<<(lsf);
        framesize = framesize + padding - 4;
        break; 
     default:
        return 0;               /* "Sorry, unknown layer type" */
    }
    return framesize;
}

static void
read_mp3(BUF* buf, M* m)
{
    unsigned long head, fs;
    while (buf->length > 4) {
        head = get_b32(buf, 0);
        if ((head & 0xFFFFFF00UL) == 0x54414700UL) {
            output_copy(m, buf, 128);
        } else {
            fs = decode_header(head);
            if (! fs)
                break;
            output_copy(m, buf, fs + 4);
        }
    }
}    

int
recog_mp3(M* m, BUF* b, size_t len)
{
    unsigned long head, fs;
    size_t        pos;
    
    if (len < 1000)
        return 0;

    pos = 0;
    while (pos < len - 4) {
        head = get_b32(b, pos);
        fs   = decode_header(head);
        if (!fs)
            break;
        if (pos + fs + 4 >= len)
            break;
        pos += fs + 4;
    }
    if (pos > 1000) {
/*         printf("mp3 length scanned = %d, buf->length = %d\n", pos, b->length); */
        match_set_ext(m, "mp3");
        match_set_reader(m, read_mp3);
        return 1;
    } else {
        return 0;
    }
}

REC mp3_recognizer = { "mp3", "\xFF", recog_mp3 };
