/**
  *  \file ringbuf.c
  */

#include <string.h>
#include <stdlib.h>
#include "ringbuf.h"
#include "util.h"

/* Initialize a buffer */
void
buf_init(BUF* buf, FILE* fp, size_t size)
{
    buf->data   = xmalloc(size);
    buf->size   = size;
    buf->pos    = 0;
    buf->fp     = fp;
    buf->length = full_read(fp, buf->data, size);
    buf->total_read = 0;
}

void
buf_free(BUF* buf)
{
    if (buf->data)
        free(buf->data);
    buf->data = 0;
}

/* Advance buffer by /by/ bytes */
void
buf_advance(BUF* buf, size_t by)
{
    size_t rem, end, now, got;
    assert(by <= buf->size);
    buf->total_read += by;
    if (by >= buf->length) {
        /* we advance over whole buffer, discarding everything in it */
        buf->pos    = 0;
        buf->length = full_read(buf->fp, buf->data, buf->size);
    } else {
        /* we advance within buffer */
        buf->pos = (buf->pos + by) % buf->size;
        buf->length -= by;

        /* attempt to re-fill buffer */
        while (1) {
            rem = buf->size - buf->length;
            if (buf->length > REFILL_THRESHOLD && rem < REFILL_THRESHOLD) {
                break;
            }
            end = (buf->pos + buf->length) % buf->size;
            now = (rem < buf->size - end) ? rem : buf->size - end;
            if (! now)
                break;
            if (! (got = full_read(buf->fp, buf->data + end, now)))
                break;
            buf->length += got;
        }
    }
    if (buf->total_read < 1024*10)
        printf("%5d bytes read\r", (int)(buf->total_read));
    else if (buf->total_read < 1024L*1024*10)
        printf("%5dk bytes read\r", (int)(buf->total_read / 1024));
    else
        printf("%5dM bytes read\r", (int)(buf->total_read / (1024L*1024)));
    fflush(stdout);
}

/* Read from buffered stream, and advance pointer */
size_t
buf_read(BUF* buf, char* data, size_t count)
{
    size_t total = 0;
    while (count) {
        size_t have = buf_avail_lin(buf);
        size_t now  = count > have ? have : count;
        if (now == 0)
            break;
        memcpy(data, buf->data + buf->pos, now);
        count -= now;
        total += now;
        data += now;
        buf_advance(buf, now);
    }
    return total;
}

/* How much data can be accesses linear? */
size_t
buf_avail_lin(BUF* buf)
{
    size_t to_end = buf->size - buf->pos;
    if (to_end < buf->length)
        return to_end;
    else
        return buf->length;
}

int
buf_streq(BUF* buf, size_t pos, const char* str)
{
    return buf_memeq(buf, pos, str, strlen(str));
}

int
buf_memeq(BUF* buf, size_t pos, const char* data, size_t length)
{
    if (length + pos > buf->length)
        return 0;
    while (length) {
        if (buf_peek(buf, pos) != *data)
            return 0;
        ++pos, ++data, --length;
    }
    return 1;
}
